<?php

namespace Teknisa\CTB\Services;

use Teknisa\CTB\Factories\Ctb12000ManutencaoLoteFactory;
use Teknisa\Libs\Util\DateUtil;
use Teknisa\Libs\Util\Environment;
use Teknisa\Libs\Util\NewCode;
use Teknisa\Model\Contabilidade\Entities\Contopectbl;
use Teknisa\Model\Contabilidade\Entities\Lancccusto;
use Teknisa\Model\Contabilidade\Entities\Lanclote;
use Teknisa\Model\Contabilidade\Entities\Lanccont;
use Teknisa\Model\Geral\Entities\Centcustfil;
use Teknisa\Model\Geral\Entities\Contctbl;
use Teknisa\Model\Geral\Entities\Filial;
use Teknisa\Model\Geral\Entities\Histpadr;
use Teknisa\Model\Geral\Entities\Centcust;

class Ctb12000ManutencaoLoteService {

    protected $entityManager;
    protected $environment;
    protected $lancamentosFactory;
    protected $newCode;
    
    const CD5_LENGTH = 5;

    public function __construct(\Doctrine\ORM\EntityManager $entityManager, Environment $environment, Ctb12000ManutencaoLoteFactory $lancamentosFactory, NewCode $newCode) {
        $this->entityManager = $entityManager;
        $this->environment = $environment;
        $this->lancamentosFactory = $lancamentosFactory;
        $this->newCode = $newCode;
    }

    /*
     * Função para verificar se uma data está entre outras duas.
     * Necessária para validar se a data do lançamento está dentro do período
     * contábil parametrizado.
     */
    private function betweenDates($compareDate, $beginDate, $endDate) {
        // É necessário truncar a data, removendo a hora, para que isso não
        // influencie na comparação.
        $beginDate = DateUtil::truncateData($beginDate);
        $endDate = DateUtil::truncateData($endDate);
        // Depois de truncada, a data final deve ser incrementada em um dia,
        // para que o último dia parametrizado também possa ser considerado.
        $endDate = DateUtil::adicionaIntervalo($endDate, 1, DateUtil::DIA);

        // Retorna true se a data não for menor que a primeira ou maior que a segunda.
        return !(DateUtil::comparaDatas($compareDate, $beginDate, DateUtil::ANTES) || DateUtil::comparaDatas($compareDate, $endDate, DateUtil::DEPOIS));
    }

    /*
     * Função para verificar se uma data de lançamento é válida.
     */
    public function checkValidDataLancamento($row) {
        $contopectblRepository = $this->entityManager->getRepository(Contopectbl::class);

        // Busca os parâmetros de contabilização da empresa selecionada.
        $contopectbl = $contopectblRepository->findOneBy(array(
            "cdempresa" => $row["CDEMPRESA"],
            "nrorg" => $this->environment->getNrOrgTrab()
        ));

        // Busca os intervalos de datas que permitem novos lançamentos contábeis.
        $dtiniclanc = $contopectbl->getDtiniclanc();
        $dtfinalanc = $contopectbl->getDtfinalanc();
        $dtiniccont = $contopectbl->getDtiniccont();
        $dtfinacont = $contopectbl->getDtfinacont();

        // Acessa a data de lançamento escolhida pelo usuário e cria o objeto do tipo DateTime.
        $row["DTLANCCONT"] = strpos($row["DTLANCCONT"], ":") === false ? $row["DTLANCCONT"] : DateUtil::truncateString($row["DTLANCCONT"]);
        $dtlanccont = DateUtil::getDataDeString($row["DTLANCCONT"], DateUtil::FORMATO_BRASILEIRO, true);

        // Valida a data de acordo com os intervalos.
        if (!$this->betweenDates($dtlanccont, $dtiniclanc, $dtfinalanc)) {
            throw new \Exception("Operação bloqueada! Data do lançamento fora do período contábil.");
        }
        if (!$this->betweenDates($dtlanccont, $dtiniccont, $dtfinacont)) {
            throw new \Exception("Operação bloqueada! Data do lançamento fora do período contábil.");
        }
    }

    /*
     * Função para excluir um lançamento.
     */
    public function deleteLanccont($row) {
        $lanccontRepository = $this->entityManager->getRepository(Lanccont::class);
        $lancloteRepository = $this->entityManager->getRepository(Lanclote::class);
        $lancamentoLote = $lanccontRepository->findOneBy(array(
            "cdempresa" => $row["CDEMPRESA"],
            "nrlanccont" => $row["NRLANCCONT"],
            "idlancinte" => "L",
            "nrorg" => $this->environment->getNrOrgTrab()
        ));
        if ($lancamentoLote !== null) {
            $lote = $lancloteRepository->findOneBy(array(
                "cdempresa" => $cabecalho["CDEMPRESA"],
                "nrlancdefi" => $partida["NRLANCCONT"],
                "nrorg" => $nrorg
            ));
            throw new \Exception("Operação bloqueada! Lançamento contábil referente a lote. Favor realizar a reversão do lote " . $lote->getCdlote() . " para as devidas alterações.");
        }

        $params = array(
            "cdempresa" => $row["CDEMPRESA"],
            "nrlanccont" => $row["NRLANCCONT"],
            "nrorg" => $this->environment->getNrOrgTrab()
        );

        $this->lancamentosFactory->deleteLanccont($params);
    }

    /*
     * Função para retornar o próximo número sequencial de um lançamento.
     */
    private function getNextSequencialLancamento($cdempresa, $nrlanccont, $nrorg) {
        $params = array(
            "CDEMPRESA" => $cdempresa,
            "NRLANCCONT" => $nrlanccont,
            "NRORG" => $nrorg
        );

        $sequencial = $this->lancamentosFactory->getNextSequencialLancamento($params)["NRSEQLANC"];
        return $sequencial === null ? "00001" : substr("0000" . ++$sequencial, -5);
    }

    /*
     * Função para retornar o número de lançamento de acordo com o cabeçalho.
     */
    private function getNrlanccont($cabecalho) {
        if ($cabecalho["lancamentoNovo"]) {
            $nrorg = $this->environment->getNrOrgTrab();
            $cabecalho["DTLANCCONT"] = strpos($cabecalho["DTLANCCONT"], ":") === false ? $cabecalho["DTLANCCONT"] : DateUtil::truncateString($cabecalho["DTLANCCONT"]);
            $dtlanccont = DateUtil::getDataDeString($cabecalho["DTLANCCONT"], DateUtil::FORMATO_BRASILEIRO, true);
            $ano = DateUtil::getAno($dtlanccont);
            $mes = DateUtil::getMes($dtlanccont);
            $dataLancamento = $ano.$mes;
            $codigoSeq = $this->newCode->getCodeByTableName("LANCCONTABIL".$cabecalho["CDEMPRESA"].$dataLancamento, $nrorg, 1, 12);
            $nrlanccont = $dataLancamento . substr($codigoSeq, 6);
            $cabecalho["NRLANCCONT"] = $nrlanccont;
        }
        else {
            $nrlanccont = $cabecalho["NRLANCCONT"];
        }
        return $nrlanccont;
    }

    /*
     * Função para retornar as partidas de um lançamento.
     */
    public function getPartidas($cdempresa, $nrlanccont) {
        // Cria os parâmetros do filtro a ser utilizado na Factory.
        $params = array(
            "CDEMPRESA" => $cdempresa,
            "NRLANCCONT" => $nrlanccont,
            "NRORG" => $this->environment->getNrOrgTrab()
        );
        return $this->lancamentosFactory->getPartidas($params);
    }

    /*
     * Função para retornar os rateios de um lançamento.
     */
    public function getRateios($cdempresa, $nrlanccont) {
        // Cria os parâmetros do filtro a ser utilizado na Factory.
        $params = array(
            "CDEMPRESA" => $cdempresa,
            "NRLANCCONT" => $nrlanccont,
            "NRORG" => $this->environment->getNrOrgTrab()
        );
        return $this->lancamentosFactory->getRateios($params);
    }
    
        /**
     * Formata Data
     * @param $data
     * @return \DateTime|null|\Setforms2\Shared\ORM\DateTime|string
     */
    public function formataData($data){
        if (!$data)
            return null;

        if (strlen($data) == 19)
            $data = DateUtil::truncateString($data);
        $data = DateUtil::getDataDeString($data);
        $data = DateUtil::truncateData($data);

        return $data;
    }
    
    private function formatValue($value, $columnName){
        switch($columnName){
            case $columnName == 'NRSEQLANLOTE':
                $value = self::leftZeroFill($value, self::CD5_LENGTH);
                break;
            default: 
                $value = strval($value);
        }
        return $value;
    }
    
    private function leftZeroFill($string, $length){
        $string = str_pad(strval($string), $length, '0', STR_PAD_LEFT); 
        return $string;
    }

    /*
     * Função que valida um lançamento e salva os dados enviados.
     */
    public function saveLanclote($cabecalho, $lancamentos, $rateios, $modoAdd){
        $connection = $this->entityManager->getConnection();
        $connection->beginTransaction();
        try {
            
            
            // Busca Dados Cabeçalho
            $documento = $cabecalho['NRDOCULOTE'];
            $data      = $cabecalho['DTLANCLOTE'];
            
            $cdempresa = $lancamentos[0]['CDEMPRESA'];
            $cdlote = $lancamentos[0]['CDLOTE'];
            $nrdoculote = $lancamentos[0]['NRDOCULOTE'];
            $dtlanclote = $this->formataData($lancamentos[0]['DTLANCLOTE']);
            $nrorg = $this->environment->getNrOrgTrab();
            $nrlanclote = $lancamentos[0]['NRLANCLOTE'];
            
            $newNrlanclote = $this->geraNovoCodigo('LANCLOTE'.$cdempresa.$cdlote, $nrorg, 1, 6);
            
            /** ROTINA DE LANÇAMENTOS - INICIO */
        
            // Busca Lançamentos Anteriores
            $lancamentos_old = $this->lancamentosFactory->buscaLanclote_old($cdempresa, $cdlote, $nrlanclote);

            // Exclui Lançamento Anteriores
            if($lancamentos_old)
                foreach($lancamentos_old as $lancamento_old)
                    $this->entityManager->remove($lancamento_old);
            $this->entityManager->flush();
            
            // Insere Novos Lançamentos
            foreach($lancamentos as $lancamento){
                $nrorg = $this->environment->getNrOrgTrab();
                $nrseqlanlote  = $lancamento['NRSEQLANLOTE'];
                $nrseqlanlote = self::formatValue($nrseqlanlote, 'NRSEQLANLOTE');
                $cdempresa = $lancamento['CDEMPRESA'];
                $cdlote = $lancamento['CDLOTE'];
                $nrdoculote = $lancamento['NRDOCULOTE'];
                $dtlanclote = $this->formataData($lancamento['DTLANCLOTE']);
                $cdcontctbl = $lancamento['CDCONTCTBL'];
                $cdhistpadr = $lancamento['CDHISTPADR'];
                $valorabs = $lancamento['VRLANCLOTE'];
                // Passando de string para float, para que o campo receba o - junto com o valor
                $valorabs = number_format(str_replace(',', '.', $valorabs), 2, ',', '.');
                $valorabs = str_replace('.', '',$valorabs);
                $valorabs = floatval(str_replace(',','.',$valorabs));
                //$idlanccome = $lancamento['IDLANCCOME'];
                $cdcontcome = $lancamento['CDLLOTCOME'];
                $iddebcre = $lancamento['IDLLOTCOME'];
                $dshistlanc = $lancamento['DSHISTLOTE'];
                
                if($modoAdd){
                    $nrlanclote = $newNrlanclote;
                }
                
                
                $dataatual = DateUtil::getDataAtual();
                $lancamento = $this->lancamentosFactory->insereLanclote($nrorg, $nrseqlanlote, $nrlanclote, $cdempresa, $cdlote, $nrdoculote, $dtlanclote, $cdcontctbl,
                $cdhistpadr, $valorabs, $cdcontcome, $iddebcre, $dshistlanc,$dataatual);
                $this->entityManager->persist($lancamento);
            }
            $this->entityManager->flush();
            
            /** ROTINA DE LANÇAMENTOS - FINAL */
            
            /** ROTINA DE RATEIOS - INICIO */
            
            foreach($rateios as $rateio){
                
                $nrorg = $this->environment->getNrOrgTrab();
                
                $cdfilial  = $rateio['CDFILIAL'];
                $cdcentcust = $rateio['CDCENTCUST'];
                $vrlanclote = $rateio['VRLANCLOTE'];
                $dtcompcusto = $this->formataData($rateio['DTCOMPCUSTO']);
                $cdempresa = $rateio['CDEMPRESA'];
                $cdlote = $rateio['CDLOTE'];
                $nrseqlanlote  = $rateio['NRSEQLANLOTE'];
                $nrseqlanlote = self::formatValue($nrseqlanlote, 'NRSEQLANLOTE');
                
                //$nrlanclote = $newNrlanclote;
                
                $rateio = $this->lancamentosFactory->insereRateio($nrorg, $cdfilial, $cdcentcust, $vrlanclote, $dtcompcusto, $cdempresa, $cdlote, $nrlanclote, $nrseqlanlote);
                
                $this->entityManager->persist($rateio);
            }
            $this->entityManager->flush();
            
            /** ROTINA DE RATEIOS - FINAL */
            
            $connection->commit();
            
            //Atualiza a capa do Lote
            $paramUpdate = array(
                "CDEMPRESA" => $cdempresa,
                "CDLOTE" => $cdlote,
                "IDLOTE" => 'D'
            );
            
            $this->lancamentosFactory->atualizaCapaLote($paramUpdate);
            
        } catch (\Exception $e) {
            $connection->rollBack();
            throw $e;
        }
    }
    
    public function geraNovoCodigo(
        $cdcontador, $nrorg, $sequencial, $tamanho
    ){
        try {
            /** @var Novocodigo $novocodigo */
            $novocodigo = $this->newCode->getCodeByTableName($cdcontador, $nrorg, $sequencial, $tamanho);
            return $novocodigo;
        } catch (\Exception $e) {
            throw $e;
            return null;
        }
    }
    
    public function deleteDadosLanc($row) {
        $params = array(
            "cdempresa" => $row["CDEMPRESA"],
            "cdlote" => $row["CDLOTE"],
            "nrlanclote" => $row["NRLANCLOTE"],
            "nrseqlanlote" => $row["NRSEQLANLOTE"],
            "nrorg" => $this->environment->getNrOrgTrab()
        );

        $this->lancamentosFactory->deleteDadosLanc($params);
    }
    
    public function deleteTodosDadosLanc($row) {
        $params = array(
            "cdempresa" => $row["CDEMPRESA"],
            "cdlote" => $row["CDLOTE"],
            "nrlanclote" => $row["NRLANCLOTE"],
            "nrorg" => $this->environment->getNrOrgTrab()
        );

        $this->lancamentosFactory->deleteTodosDadosLanc($params);
    }

    /*
     * Função que valida os dados gerais de um lançamento.
     */
    private function validateCabecalho($cabecalho) {
        if (empty($cabecalho["CDDOCUCONT"])) {
            throw new \Exception("Documento deve ser informado.");
        }
        if (empty($cabecalho["DTLANCCONT"])) {
            throw new \Exception("Data do lançamento contábil deve ser informada.");
        }
        if (empty($cabecalho["CDEMPRESA"])) {
            throw new \Exception("Empresa deve ser informada.");
        }
        if (empty($cabecalho["NRLANCCONT"]) && empty($cabecalho["lancamentoNovo"])) {
            throw new \Exception("Número do lançamento deve ser informado.");
        }

        $contopectblRepository = $this->entityManager->getRepository(Contopectbl::class);
        $contopectbl = $contopectblRepository->findOneBy(array(
            "cdempresa" => $cabecalho["CDEMPRESA"],
            "nrorg" => $this->environment->getNrOrgTrab()
        ));
        if ($contopectbl === null) {
            throw new \Exception("Não há parâmetro contábil registrado para a empresa selecionda.");
        }


        $cabecalho["DTLANCCONT"] = strpos($cabecalho["DTLANCCONT"], ":") === false ? $cabecalho["DTLANCCONT"] :  DateUtil::truncateString($cabecalho["DTLANCCONT"]);
        $dtlanccont = DateUtil::getDataDeString($cabecalho["DTLANCCONT"], DateUtil::FORMATO_BRASILEIRO, true);
        $dtiniclanc = $contopectbl->getDtiniclanc();
        $dtfinalanc = $contopectbl->getDtfinalanc();
        $dtiniccont = $contopectbl->getDtiniccont();
        $dtfinacont = $contopectbl->getDtfinacont();
        if (!$this->betweenDates($dtlanccont, $dtiniclanc, $dtfinalanc)) {
            throw new \Exception("Operação bloqueada! Data do lançamento fora do período contábil.");
        }
        if (!$this->betweenDates($dtlanccont, $dtiniccont, $dtfinacont)) {
            throw new \Exception("Operação bloqueada! Data do lançamento fora do período contábil.");
        }

        return $contopectbl;
    }

    /*
     * Função que valida os dados gerais de um lançamento.
     */
    private function validatePartidas(Array $partidas, Contopectbl $contopectbl, $cabecalho) {
        if (empty($partidas)) {
            throw new \Exception("Operação bloqueada! Nenhuma partida foi informada para o lançamento.");
        }

        $contctblRepository = $this->entityManager->getRepository(Contctbl::class);
        $lancloteRepository = $this->entityManager->getRepository(Lanclote::class);
        $nrorg = $this->environment->getNrOrgTrab();

        $utilizaRateios = $contopectbl->getIdcontgere() == 'S';
        $utilizaCodigoComercial = $contopectbl->getIdcontcome() == 'S';

        $creditos = 0;
        $debitos = 0;
        $valorTotal = 0;

        foreach($partidas as $partida) {
            if (empty($partida["CDCONTREDU"])) {
                throw new \Exception("Código Reduzido deve ser informado.");
            }
            if (empty($partida["CDCONTCTBL"])) {
                throw new \Exception("Conta Contábil deve ser informada.");
            }
            if (empty($partida["CDHISTPADR"])) {
                throw new \Exception("Código do Histórico Padrão deve ser informado.");
            }
            if (empty($partida["DSHISTLANC"])) {
                throw new \Exception("Descrição do Lançamento deve ser informada.");
            }
            if (empty($partida["VRLANCAMENT"])) {
                throw new \Exception("Valor da partida deve ser informado.");
            }
            if ($partida["VRLANCAMENT"] == 0) {
                throw new \Exception("Valor da partida deve ser diferente de zero.");
            }
            if (!empty($partida["IDLANCINTE"]) && $partida["IDLANCINTE"] == 'L') {
                $lote = $lancloteRepository->findOneBy(array(
                    "cdempresa" => $cabecalho["CDEMPRESA"],
                    "nrlancdefi" => $partida["NRLANCCONT"],
                    "nrorg" => $nrorg
                ));
                throw new \Exception("Operação bloqueada! Lançamento contábil referente a lote. Favor realizar a reversão do lote " . $lote->getCdlote() . " para as devidas alterações.");
            }

            $contctbl = $contctblRepository->findOneBy(array(
                "cdcontctbl" => $partida["CDCONTCTBL"],
                "nrorg" => $nrorg
            ));
            if ($contctbl === null) {
                throw new \Exception("Conta contábil inválida.");
            }
            if ($contctbl->getIdtipocont() != "A") {
                throw new \Exception("Conta contábil deve ser do tipo analítica.");
            }
            if ($utilizaCodigoComercial && $contctbl->getIdtipocome() !== null) {
                if (empty($partida["CDCONTCOME"])) {
                    throw new \Exception("Código Comercial deve ser informado.");
                }
                if (empty($partida["IDLANCCOME"])) {
                    throw new \Exception("Identificador do Código Comercial deve ser informado.");
                }
                if ($partida["IDLANCCOME"] != 'B' && $partida["IDLANCCOME"] != 'C' && $partida["IDLANCCOME"] != 'F') {
                    throw new \Exception("Identificador do Código Comercial inválido.");
                }
            }
            if ($utilizaRateios && ($contctbl->getIdcontccus() == 'R' || $contctbl->getIdcontccus() == 'G')) {
                if (empty($partida["rateios"])) {
                    throw new \Exception("Rateio deve ser informado para as contas do tipo Gerencial ou Oficial/Gerencial.");
                }

                $valorRateios = $this->validateRateios($partida["rateios"], $contopectbl, $cabecalho);
                if ($valorRateios != $partida["VRLANCAMENT"]) {
                    throw new \Exception("A soma dos rateios deve ser igual ao valor da partida.");
                }
            }

            if ($partida["VRLANCAMENT"] > 0) {
                $creditos += $partida["VRLANCAMENT"];
            }
            else {
                $debitos += $partida["VRLANCAMENT"];
            }
            $valorTotal += $partida["VRLANCAMENT"];

            // Correção de erros de operações binárias, que podem resultar em
            // valores com 15+ casas decimais.
            $valorTotal = round($valorTotal, 2);
        }

        if ($creditos == 0) {
            throw new \Exception("Partida de crédito deve ser realizada.");
        }
        if ($debitos == 0) {
            throw new \Exception("Partida de débito deve ser realizada.");
        }
        if ($valorTotal != 0) {
            throw new \Exception("Valor total de crédito difere do total de débito.");
        }
    }

    /*
     * Função que valida os rateios de uma partida.
     */
    private function validateRateios(Array $rateios, Contopectbl $contopectbl, $cabecalho) {
        $centcustfilRepository = $this->entityManager->getRepository(Centcustfil::class);
        $filialRepository = $this->entityManager->getRepository(Filial::class);
        $nrorg = $this->environment->getNrOrgTrab();

        $valorTotal = 0;
        $filialCentcust = array();

        foreach($rateios as $rateio) {
            if (empty($rateio["CDFILIAL"])) {
                throw new \Exception("Unidade do rateio deve ser informada.");
            }
            if (empty($rateio["CDCENTCUST"])) {
                throw new \Exception("Centro de Custo do rateio deve ser informado.");
            }
            if (empty($rateio["DTCOMPCUSTO"])) {
                throw new \Exception("Data da Competência do rateio deve ser informado.");
            }

            $filial = $filialRepository->findOneBy(array(
                "cdfilial" => $rateio["CDFILIAL"],
                "nrorg" => $nrorg
            ));
            if ($filial === null) {
                throw new \Exception("Unidade selecionada inválida.");
            }
            if ($filial->getCdempresa() != $cabecalho["CDEMPRESA"]) {
                throw new \Exception("Unidade não relacionada à empresa do lançamento contábil.");
            }

            $cdfilial = $filial->getCdfilial();
            if (!array_key_exists($cdfilial, $filialCentcust)) {
                $filialCentcust[$cdfilial] = array();
            }

            $centcustfil = $centcustfilRepository->findOneBy(array(
                "cdfilial" => $rateio["CDFILIAL"],
                "cdcentcust" => $rateio["CDCENTCUST"],
                "nrorg" => $nrorg
            ));
            if ($centcustfil === null) {
                throw new \Exception("Centro de Custo selecionado inválido.");
            }
            if ($centcustfil->getCdfilial() != $cdfilial) {
                throw new \Exception("Centro de Custo não relacionada à unidade do rateio.");
            }

            $cdcentcust = $centcustfil->getCdcentcust();
            if (in_array($centcustfil->getCdcentcust(), $filialCentcust[$cdfilial])) {
                throw new \Exception("Rateio já informado para a unidade e centro de custo informados.");
            }
            $filialCentcust[$cdfilial][] = $cdcentcust;

            if($rateio["VRLANCCUST"] == 0) {
                throw new \Exception("Valor do rateio deve ser diferente de zero.");
            }
            $valorTotal += $rateio["VRLANCCUST"];
        }

        return $valorTotal;
    }
    
     /*
     * Função que 
     */
    public function rastreioContabil($row) {
        try{
           $rastreio = $this->lancamentosFactory->rastreioContabil(
               $row['0'],
               $row['1']
            );
        } catch (\Exception $e) {
            throw $e;
        }
        return $rastreio[0];
    }
    
    public function conferirLote($lotes){
        try {
            $semError = true;
            foreach($lotes as $lote){
                $cdempresa = $lote['CDEMPRESA'];
                $cdlote = $lote['CDLOTE'];
                
                $lote = $this->lancamentosFactory->execProcCONFERE_ERROS_LOTE($cdempresa, $cdlote);
                if($lote == 'false')
                    $semError = false;
            }
            return $semError;
        } catch (\Exception $e) {
            throw $e;
        }
    }
    
    // se tiver conferido habilita
    public function liberarLote($lotes){
        try {
            foreach($lotes as $lote){
                $cdempresa = $lote['CDEMPRESA'];
                $cdlote = $lote['CDLOTE'];
                
                $nrlanc = null;
                $cdoperador = $this->environment->getCdOperador();
                
                $lote = $this->lancamentosFactory->execProcLIBERA_LOTE($cdempresa, $cdlote, $nrlanc, $cdoperador);
            }
         
        } catch (\Exception $e) {
            throw $e;
        }
    }
    
    // se tiver liberado habilita
    public function reverterLote($lotes){
        try {
            foreach($lotes as $lote){
                $cdempresa = $lote['CDEMPRESA'];
                $cdlote = $lote['CDLOTE'];
                
                $lote = $this->lancamentosFactory->execProcREVERTE_LOTE($cdempresa, $cdlote);

            }
         
        } catch (\Exception $e) {
            throw $e;
        }
    }
    
    
    public function readFile($row){
        try {
            $file = $this->decodeFile(substr($row["UPLOADFILE"][0]["b64File"], 23));
            $stringFile = utf8_encode($file);
            $arrayContentFile = $this->returnStringInlines($stringFile);
            
            if($row["IDMODELO"] == 'P'){
                $readFile = $this->readFileModelPattern($arrayContentFile);
                $validations = $this->validationsContent($readFile,$row["CDEMPRESA"],$row["CDLOTE"]);
                if($validations){
                   return $validations; 
                } else{
                    $this->insertImportLote($readFile,$row["CDEMPRESA"],$row["CDLOTE"]);
                    $lote = ["SUCESSO"];
                    return $lote;
                }
                
            } else {
               // $readFile = $this->readFileModelMetadados($arrayContentFile);
            }
            
        } catch (\Exception $e) {
            throw $e;
        }
    }
    
    /**
     * Essa função tem como finalidade receber uma string codificada em base 64 e
     * descidificala para ser usada nas funções dessa classe.
     *
     * @param  string $encodedFile codificada em base64
     * @return string              decodificada
     */
    public function decodeFile($encodedFile) {
        $file = base64_decode($encodedFile);
        return $file;
    }
    
    
    /** ------------------------------- LEIAUTE PADRÃO -------------------------------
          001 | 06 posições | Sequencial
          007 | 08 posições | Data de Lançamento AAAAMMDD
          015 | 04 posições | Centro de custo a débito
          019 | 04 posições | Filial a débito
          023 | 04 posições | Centro de custo a débito
          027 | 04 posições | Filial a crédito
          031 | 08 posições | Número do documento
          039 | 20 posições | Conta Contábil a Débito
          059 | 20 posições | Conta Contábil a Crédito
          079 | 04 posições | Código do Histórico Padrão
          083 |200 posições | Descrição Histórico Padrão
          283 | 16 posições | Valor do lançamento
          299 | 01 posições | Identificador Comercial
          300 | 21 posições | Código Comercial
    */
    public function readFileModelPattern($contentFile) {
        $data = array();
        foreach ($contentFile as $line) {
            if ($line != ""){
                
                $nrLancLote = trim(substr($line, 0, 6));
                
                //Monta a data e movimentação contida na linha
                $dtSemFormatacao = trim(substr($line, 6, 8));
                
                $dia = trim(substr($dtSemFormatacao, 6, 2));
                $mes = trim(substr($dtSemFormatacao, 4, 2));
                $ano = trim(substr($dtSemFormatacao, 0, 4));
                $dtLanc = $dia.'/'.$mes.'/'.$ano;

                $centCustDeb = trim(substr($line, 14, 4));
                
                $cdFilialDeb = trim(substr($line, 18, 4));
                
                $cdCentCust = trim(substr($line, 22, 4));
                
                $cdFilial = trim(substr($line, 26, 4));
                
                $nrDoc = trim(substr($line, 30, 8));
                
                $cdContCtblDeb = trim(substr($line, 38, 20));
                
                $cdContCtbl = trim(substr($line, 58, 20));
                
                $cdContHistPadr = trim(substr($line, 78, 4));
                
                $dsHistLote = trim(substr($line, 82, 200));
                
                $vrLanInc = trim(substr($line, 282, 16));
                
                $idComercial = trim(substr($line, 298, 1));
                
                $cdComercial = trim(substr($line, 299, 21));
                
                $lines = ["NRLANCLOTE" => $nrLancLote, 
                          "DTLANCLOTE" => $dtLanc,
                          "CDCENTCUSTDEB" => $centCustDeb,
                          "CDFILIALDEB" => $cdFilialDeb,
                          "CDCENTCUST" => $cdCentCust,
                          "CDFILIAL" => $cdFilial,
                          "NRDOCU" => $nrDoc,
                          "CDCONTCTBLDEB" => $cdContCtblDeb,
                          "CDCONTCTBL" => $cdContCtbl,
                          "CDCONTHISTPADR" => $cdContHistPadr,
                          "DSHISTLOTE" => $dsHistLote,
                          "VRLANINC" => $vrLanInc,
                          "IDCOMERCIAL" => $idComercial,
                          "CDCOMERCIAL" => $cdComercial];
                          
                array_push($data,$lines);
            }
        }
        return $data;
    }
    
    
    /**{------------------------------- LEIAUTE METADADOS -------------------------------
          001 a 004 | 04 posições | Código da Empresa (NÃO UTILIZADO, É INFORMADO EMPRESA NA TELA)
          005 a 008 | 04 posições | Código do Estabelecimento, Código da Unidade ou Branco (NÃO UTILIZADO)
          009 a 012 | 04 posições | Código da Classificação contábil (Código da Filial)
          013 a 032 | 20 posições | Classificação Contábil Estendidada (NÃO UTILIZADO)
          033 a 035 | 03 posições | Item Contábil (NÃO UTILIZADO)
          036 a 075 | 40 posições | Descrição do Item Contábil (NÃO UTILIZADO)
          076 a 076 | 01 posições | Típo de Item: 1- Contábil ou 2 -Estático  (NÃO UTILIZADO)
          077 a 084 | 08 posições | Data do Lançamento no formato AAAAMMDD
          085 a 104 | 20 posições | Conta de débito
          105 a 124 | 20 posições | Conta de crédito
          125 a 128 | 04 posições | Código do Histórico Padrão
          129 a 129 | 01 posições | Se valor do lançamento negativo ( - ), senão branco -> Sinal colocado diretamente no servidor, na contrapartida.
          130 a 144 | 15 posições | Valor da partida (sem seperador decimal) -> Para pegar casa decimal foi dividido o valor por 10
          145 a 152 | 08 posições | Classificação Contábil c/ mais de 4 dígitos (NÃO UTILIZADO)
          153 a 172 | 20 posições | Código Auxiliar do Estabelecimento (NÃO UTILIZADO)
         }
    */
    public function readFileModelMetadados($contentFile) {

         foreach ($contentFile as $line) {
            if ($line != ""){
                
                $cdFilial =  trim(substr($line, 5, 4));
                
                if(empty($cdFilial)){
                    $cdFilial =  trim(substr($line, 9, 4));
                }
                
                //Monta a data e movimentação contida na linha
                $dia = trim(substr($line, 83, 2));
                $mes = trim(substr($line, 81, 2));
                $ano = trim(substr($line, 77, 4));
                $dtLanc = $dia.'/'.$mes.'/'.$ano;
                
                $cdContDeb = trim(substr($line, 85, 20));
                $contctblRepository = $this->entityManager->getRepository(Contctbl::class);
                $filialRepository = $this->entityManager->getRepository(Filial::class);
                $nrorg = $this->environment->getNrOrgTrab();

                $contctbl = $contctblRepository->findOneBy(array(
                    "cdcontctbl" => $cdContDeb,
                    "nrorg" => $nrorg
                ));
                
                $idcontccus = $contctbl->getIdcontccus();
                if($idcontccus != 'O'){
                    if($idcontccus == 'G' or $idcontccus == 'R'){
                        if(empty($cdFilial)){
                            throw new \Exception("Filial Não Informada");
                        }
                    }
                }
                
                $cdContCred = trim(substr($line, 105, 20));
                
                if($idcontccus != 'O'){
                    if($idcontccus == 'G' or $idcontccus == 'R'){
                        if(empty($cdFilial)){
                            throw new \Exception("Filial Não Informada");
                        }
                    }
                }
                
                // Faz Validação da Filial/ Centro de Custo
                if(!empty($cdFilial)){
                    $filial = $filialRepository->getCdfilial();
                    if(empty($filial)){
                        throw new \Exception("Filial Não Enconrtada");
                    }
                }
                
                $cdHistPadr = trim(substr($line, 125, 4));
                if($cdHistPadr == ''){
                    //$cdHistPadr = ''; Aqui receve uma descrição padrão que não sei o que colocar
                }
                
                
                $vrLancLote = trim(substr($line, 130, 15));
                $vrLancLote = $vrLancLote * 100;
                
                $lines = ["CDFILIAL" => $cdFilial,
                          "DTLANCLOTE" => $dtLanc,
                          "CDCONTDEB" => $contctbl,
                          "CDCONTCRED" => $cdContCred,
                          "CDHISTPADR" => $cdHistPadr,
                          "VRLANCLOTE" => $vrLancLote];
                array_push($data,$lines);
            }
            
         }
         return $data;
    }
    
    
    
    /**
     * Essa função tem como finalidade receber uma string que contém quebras de linhas,
     * e retornar em um array as linhas, onde cada posição do array contém uma linha da
     * string lida.
     *
     * @param $file | string completa com quebras de linha
     *
     * @return $file | array de strings, onde cada posição é uma linha
     */
    public function returnStringInlines($file) {
        $file = preg_split('/\R/', $file);
        return $file;
    }
    /**
     * Essa função tem como finalidade validar as informações do lote
     * Campos: DATA/CONTA-CONTABIL/CENTRO DE CUSTO/FILIAL/LANCAMENTO
     * 
    */
    public function validationsContent($content, $empresa, $lote) {
        $contador = 0;
        $erros = array();
        // Itera sobre os lançamentos para fazer as validações
        foreach($content as $contents){
            $contador++;
            
             $paramLote = array(
                "CDEMPRESA" => $empresa,
                "CDLOTE" => $lote,
            );
            
            //Verifica se o Lote já existe
            $loteExist = $this->lancamentosFactory->buscaLote($paramLote);
            if($loteExist){
                array_push($erros, [
                    "CDLINHA" => $contador,
                    "DESCERRO" => "O Lote:"." ".$lote." "."já existe"
                ]);
            }
            //Verifica se existe e se já existe um lançamento
            if($contents["NRLANCLOTE"] != ""){
                $lanExist = $this->lancamentosFactory->buscaLanclote_old($empresa, $lote, $contents["NRLANCLOTE"]);
                if($lanExist) {
                    array_push($erros, [
                        "CDLINHA" => $contador,
                        "DESCERRO" => "O Lançamento:"." ".$contents["NRLANCLOTE"]." "."já existe"
                    ]);
                }
            } else {
                array_push($erros, [
                    "CDLINHA" => $contador,
                    "DESCERRO" => "Numero do Lançamento não Informado no Arquivo"
                ]);
            }
            
            // valida a data passada no arquivo é valida ou se foi informada no arquivo
            if($contents["DTLANCLOTE"]){
              $data = explode( '/', $contents["DTLANCLOTE"]);
              $validacao = checkdate($data[1], $data[0], $data[2]);
              if($validacao == false){
                    array_push($erros, [
                    "CDLINHA" => $contador,
                    "DESCERRO" => "A data"." ".$contents["DTLANCLOTE"]." "."é inválida"
                ]);
              }  
            } else {
                array_push($erros, [
                    "CDLINHA" => $contador,
                    "DESCERRO" => "Data não Informada no Arquivo"
                ]);
            }
            
            //Valida se o código do centro de custo foi informado e se existe
            //parametro opicional no arquivo
            if($contents["CDCENTCUSTDEB"] != ""){
                $centcustRepository = $this->entityManager->getRepository(Centcust::class);
                $nrorg = $this->environment->getNrOrgTrab();
                $centCust = $centcustRepository->findOneBy(array(
                    "cdcentcust" => $contents["CDCENTCUSTDEB"],
                    "nrorg" => $nrorg
                ));
                
                if(!$centCust) {
                    array_push($erros, [
                        "CDLINHA" => $contador,
                        "DESCERRO" => "Código do Centro de Custo a Débito não Cadastrado"
                    ]);
                }
            }
            
            
            //Valida se o código da filial a débito foi informado e se existe
            // Parametro opcional no arquivo
            if($contents["CDFILIALDEB"] != ""){
                $filialRepository = $this->entityManager->getRepository(Filial::class);
                $nrorg = $this->environment->getNrOrgTrab();
                
                $filial = $filialRepository->findOneBy(array(
                    "cdfilial" => $contents["CDFILIALDEB"],
                    "nrorg" => $nrorg
                ));
                
                if(!$filial){
                    array_push($erros, [
                        "CDLINHA" => $contador,
                        "DESCERRO" => "Código da Filial a Débito não Cadastrado"
                    ]);  
                }
            }
            
            
            //Valida se o código do centro de custo foi informado e se existe
            //Parametro opcional no arquivo
            if($contents["CDCENTCUST"] != ""){
                $centcustRepository = $this->entityManager->getRepository(Centcust::class);
                $nrorg = $this->environment->getNrOrgTrab();
                $centCust = $centcustRepository->findOneBy(array(
                    "cdcentcust" => $contents["CDCENTCUST"],
                    "nrorg" => $nrorg
                ));
                
                if(!$centCust) {
                    array_push($erros, [
                        "CDLINHA" => $contador,
                        "DESCERRO" => "Código do Centro de Custo a Crédito não Cadastrado"
                    ]);
                }
            }
            
            //Valida se o código da filial a débito foi informado e se existe
            // Parametro opcional no arquivo
            
            if($contents["CDFILIAL"] != ""){
                $filialRepository = $this->entityManager->getRepository(Filial::class);
                $nrorg = $this->environment->getNrOrgTrab();
                
                $filial = $filialRepository->findOneBy(array(
                    "cdfilial" => $contents["CDFILIAL"],
                    "nrorg" => $nrorg
                ));
                
                if(!$filial){
                    array_push($erros, [
                        "CDLINHA" => $contador,
                        "DESCERRO" => "Código da Filial a Crédito não Cadastrado"
                    ]);  
                }
            }
        
            //valida se foi passado a conta contabil de débito foi e se esta cadastrada
            if($contents["CDCONTCTBLDEB"] != ""){
                $contctblRepository = $this->entityManager->getRepository(Contctbl::class);
                $nrorg = $this->environment->getNrOrgTrab();
                $contctbl = $contctblRepository->findOneBy(array(
                    "cdcontctbl" => $contents["CDCONTCTBLDEB"],
                    "nrorg" => $nrorg
                ));
                
                if(!$contctbl){
                    array_push($erros, [
                        "CDLINHA" => $contador,
                        "DESCERRO" => "Conta Contábil de Débito não Cadastrado"
                    ]); 
                }
            } else {
                array_push($erros, [
                    "CDLINHA" => $contador,
                    "DESCERRO" => "Conta Contábil de Débito não Informado no Arquivo"
                ]); 
            }
            
            //valida se foi passado a conta contabil de credito e se esta cadastrada
            if($contents["CDCONTCTBL"] != ""){
                $contctblRepository = $this->entityManager->getRepository(Contctbl::class);
                $nrorg = $this->environment->getNrOrgTrab();
                $contctbl = $contctblRepository->findOneBy(array(
                "cdcontctbl" => $contents["CDCONTCTBL"],
                    "nrorg" => $nrorg
                ));
                
                if(!$contctbl){
                   array_push($erros, [
                        "CDLINHA" => $contador,
                        "DESCERRO" => "Conta Contábil de Crédito não Cadastrado"
                    ]); 
                }
            } else {
                array_push($erros, [
                    "CDLINHA" => $contador,
                    "DESCERRO" => "Conta Contábil de Crédito não Informado no Arquivo"
                ]); 
            }
            
            //valida se foi passado a cod historico padrao e se esta cadastrada
            if($contents["CDCONTHISTPADR"] != ""){
                $cdHistoricoPadrao =  $this->entityManager->getRepository(Histpadr::class);
                $nrorg = $this->environment->getNrOrgTrab();
                $cdHistPadr = $cdHistoricoPadrao->findOneBy(array(
                       "cdhistpadr" => $contents["CDCONTHISTPADR"],
                       "nrorg" => $nrorg
                ));
                
                if(!$cdHistPadr){
                    array_push($erros, [
                        "CDLINHA" => $contador,
                        "DESCERRO" => "Código de Histórico Padrão não Cadastrado"
                    ]); 
                }
            }
            
        }
        return $erros;
    }
    
    public function insertImportLote($content, $empresa, $lote){
        $connection = $this->entityManager->getConnection();
        $connection->beginTransaction();
        try{
            $rateio = array();
            $lancamentoContabil = array();
            $nrorg = $this->environment->getNrOrgTrab();
            $sequencial = 0;
            $loteAntigo = $content[0]["NRLANCLOTE"];
            $dataAtual = DateUtil::getDataAtual();
            $capaLote = array(
                "CDEMPRESA" => $empresa,
                "CDLOTE" => $lote,
                "DTINICLOTE" => $dataAtual->format('d/m/Y'),
                "DTFINALOTE" => $dataAtual->format('d/m/Y'),
                "DSOBSELOTE" => "Lote Importado",
                "IDLOTE" => "D"
            );
            
            $this->lancamentosFactory->insertCapaLote($capaLote);
            
            // Monta o Array para inserir os lançamento
            foreach($content as $contents){
                
                $nrlanclote = $contents["NRLANCLOTE"];
                
                if($loteAntigo == $nrlanclote){
                    $sequencial++;
                } else{
                    $loteAntigo = $nrlanclote;
                    $sequencial = 1;
                }
                
                $nrseqlanlote = str_pad( $sequencial, 5 , '0' , STR_PAD_LEFT);

                $cdempresa = $empresa;
                $cdlote = $lote;
                $nrdoculote = $contents["NRDOCU"];
                $dtlanclote = $contents["DTLANCLOTE"];
                $dtlanclote = $this->formataData($dtlanclote);
                $cdcontcome = $contents["CDCOMERCIAL"];
                $iddebcre = $contents["IDCOMERCIAL"];
                $dshistlanc = $contents["DSHISTLOTE"];
                $cdhistpadr = $contents["CDCONTHISTPADR"];
                
                $valorabs = number_format(str_replace(',', '.', $contents["VRLANINC"]), 2, ',', '.');
                $valorabs = str_replace('.', '',$valorabs);
                $valorabs = floatval(str_replace(',','.',$valorabs));
                
                //Monta array para inserir o Rateio
                if($contents["CDCENTCUSTDEB"] != "" && $contents["CDFILIALDEB"] != ""){
                    array_push($rateio,[
                        "NRORG" => $nrorg,
                        "CDFILIAL" => $contents["CDFILIALDEB"],
                        "CDCENTCUST" => $contents["CDCENTCUSTDEB"],
                        "VRLANCLOTE" => $valorabs * (-1),
                        "DTCOMPCUSTO"=> $dtlanclote,
                        "CDEMPRESA" => $empresa,
                        "CDLOTE" => $lote,
                        "NRLANCLOTE" => $nrlanclote,
                        "NRSEQLANLOTE" => str_pad($nrseqlanlote + 1, 5 , '0' , STR_PAD_LEFT)
                    ]);
                }
                if($contents["CDCENTCUST"] != "" && $contents["CDFILIAL"] != ""){
                    array_push($rateio,[
                        "NRORG" => $nrorg,
                        "CDFILIAL" => $contents["CDFILIAL"],
                        "CDCENTCUST" => $contents["CDCENTCUST"],
                        "VRLANCLOTE" => $valorabs,
                        "DTCOMPCUSTO"=> $dtlanclote,
                        "CDEMPRESA" => $empresa,
                        "CDLOTE" => $lote,
                        "NRLANCLOTE" => $nrlanclote,
                        "NRSEQLANLOTE" => $nrseqlanlote
                    ]);
                }
                
            //Desmembra a linha para inserir ao array auxiliar a partida e contra partida a ser inserida
            if($contents["CDCONTCTBL"]){
                array_push($lancamentoContabil,[
                    "NRORG" => $nrorg,
                    "NRSEQLANLOTE" => $nrseqlanlote,
                    "NRLANLOTE" => $nrlanclote,
                    "CDEMPRESA" => $cdempresa,
                    "CDLOTE" => $cdlote,
                    "NRDOCULOTE" => $nrdoculote,
                    "DTLANCLOTE" => $dtlanclote,
                    "CDCONTCTBL" => $contents["CDCONTCTBL"],
                    "CDHISTPADR" => $cdhistpadr,
                    "VALOR" => $valorabs,
                    "CDCONTCOME" => $cdcontcome,
                    "IDDEBCRE" => $iddebcre,
                    "DSHISTLANC" =>$dshistlanc,
                    "DATAATUAL" => DateUtil::getDataAtual()
                ]);
            }
             if($contents["CDCONTCTBLDEB"]){
                array_push($lancamentoContabil,[
                    "NRORG" => $nrorg,
                    "NRSEQLANLOTE" => str_pad($nrseqlanlote + 1, 5 , '0' , STR_PAD_LEFT),
                    "NRLANLOTE" => $nrlanclote,
                    "CDEMPRESA" => $cdempresa,
                    "CDLOTE" => $cdlote,
                    "NRDOCULOTE" => $nrdoculote,
                    "DTLANCLOTE" => $dtlanclote,
                    "CDCONTCTBL" => $contents["CDCONTCTBLDEB"],
                    "CDHISTPADR" => $cdhistpadr,
                    "VALOR" => $valorabs * (-1),
                    "CDCONTCOME" => $cdcontcome,
                    "IDDEBCRE" => $iddebcre,
                    "DSHISTLANC" =>$dshistlanc,
                    "DATAATUAL" => DateUtil::getDataAtual()
                ]);
            }
            }
            
            //Insere as partidas
            foreach($lancamentoContabil as $parts){
                $lancamento = $this->lancamentosFactory->insereLanclote($parts["NRORG"],$parts["NRSEQLANLOTE"], $parts["NRLANLOTE"],$parts["CDEMPRESA"],$parts["CDLOTE"],
                $parts["NRDOCULOTE"],$parts["DTLANCLOTE"],$parts["CDCONTCTBL"], $parts["CDHISTPADR"],$parts["VALOR"], $parts["CDCONTCOME"],$parts["IDDEBCRE"],$parts["DSHISTLANC"],$parts["DATAATUAL"]);
                $this->entityManager->persist($lancamento);
            }

           $this->entityManager->flush();
            $paramUpdate = array(
                "CDEMPRESA" => $empresa,
                "CDLOTE" => $lote,
                "IDLOTE" => 'D'
            );
            //Atualiza a capa do Lote
            $this->lancamentosFactory->atualizaCapaLote($paramUpdate);
            
            if($rateio){
            
              foreach($rateio as $dadosRateio){
                  $nrorg = $dadosRateio["NRORG"];
                  $cdfilial = $dadosRateio["CDFILIAL"];
                  $cdcentcust = $dadosRateio["CDCENTCUST"];
                  $vrlanclote = $dadosRateio["VRLANCLOTE"];
                  $dtcompcusto = $dadosRateio["DTCOMPCUSTO"];
                  $cdempresa = $dadosRateio["CDEMPRESA"];
                  $cdlote = $dadosRateio["CDLOTE"];
                  $nrlanclote = $dadosRateio["NRLANCLOTE"];
                  $nrseqlanlote = $dadosRateio["NRSEQLANLOTE"];
              // insere rateio  
              $rateio = $this->lancamentosFactory->insereRateio($nrorg, $cdfilial, $cdcentcust, $vrlanclote, $dtcompcusto, $cdempresa, $cdlote, $nrlanclote, $nrseqlanlote);
                
              $this->entityManager->persist($rateio);
              
              }  
              $this->entityManager->flush();
            }
        $connection->commit();
           
        }catch (\Exception $e) {
            throw $e;
        }
    }
    
}